import logging
from pkg_resources import iter_entry_points

import redis
import fakeredis
import dataset
import storagelayer
from werkzeug.local import LocalProxy
from raven.handlers.logging import SentryHandler

from memorious import settings


log = logging.getLogger(__name__)


redis_pool = redis.ConnectionPool(
    host=settings.REDIS_HOST,
    port=settings.REDIS_PORT,
    decode_responses=True
)


# set up raven for error reporting
if settings.SENTRY_DSN:
    handler = SentryHandler(settings.SENTRY_DSN)
    handler.setLevel(logging.ERROR)

# File storage layer for blobs on local file system or S3
storage = storagelayer.init(settings.ARCHIVE_TYPE,
                            path=settings.ARCHIVE_PATH,
                            aws_key_id=settings.ARCHIVE_AWS_KEY_ID,
                            aws_secret=settings.ARCHIVE_AWS_SECRET,
                            aws_region=settings.ARCHIVE_AWS_REGION,
                            bucket=settings.ARCHIVE_BUCKET)


def load_manager():
    if not hasattr(settings, '_manager'):
        from memorious.logic.manager import CrawlerManager
        settings._manager = CrawlerManager()
        if settings.CONFIG_PATH:
            settings._manager.load_path(settings.CONFIG_PATH)
    return settings._manager


def load_datastore():
    if not hasattr(settings, '_datastore'):
        if not settings.DATASTORE_URI:
            raise RuntimeError("No $MEMORIOUS_DATASTORE_URI.")
        settings._datastore = dataset.connect(settings.DATASTORE_URI)
        # Use bigint to store integers by default
        settings._datastore.types.integer = settings._datastore.types.bigint
    return settings._datastore


manager = LocalProxy(load_manager)
datastore = LocalProxy(load_datastore)


def connect_redis():
    if not settings.REDIS_HOST:
        return fakeredis.FakeRedis(decode_responses=True)
    return redis.Redis(connection_pool=redis_pool, decode_responses=True)


def load_extensions():
    for ep in iter_entry_points('memorious.plugins'):
        func = ep.load()
        func()


def init_memorious():
    load_extensions()



from memorious.model.event import Event  # noqa
from memorious.model.tag import Tag  # noqa
from memorious.model.common import Base  # noqa
from memorious.model.crawler_state import CrawlerState  # noqa
from memorious.model.crawler_run import CrawlerRun  # noqa
from memorious.model.queue import Queue  # noqa



import os
import io
import yaml
import logging
from datetime import timedelta, datetime

from memorious import settings
from memorious.model import Tag, Event, CrawlerState, CrawlerRun, Queue
from memorious.logic.stage import CrawlerStage
from memorious.task_runner import TaskRunner

log = logging.getLogger(__name__)


class Crawler(object):
    """A processing graph that constitutes a crawler."""
    SCHEDULES = {
        'daily': timedelta(days=1),
        'weekly': timedelta(weeks=1),
        'monthly': timedelta(weeks=4)
    }

    def __init__(self, manager, source_file):
        self.manager = manager
        self.source_file = source_file
        with io.open(source_file, encoding='utf-8') as fh:
            self.config_yaml = fh.read()
            self.config = yaml.load(self.config_yaml)

        self.name = os.path.basename(source_file)
        self.name = self.config.get('name', self.name)
        self.description = self.config.get('description', self.name)
        self.category = self.config.get('category', 'scrape')
        self.schedule = self.config.get('schedule')
        self.disabled = self.config.get('disabled', False)
        self.init_stage = self.config.get('init', 'init')
        self.delta = Crawler.SCHEDULES.get(self.schedule)
        self.delay = int(self.config.get('delay', 0))
        self.expire = int(self.config.get('expire', settings.EXPIRE)) * 84600
        self.stealthy = self.config.get('stealthy', False)

        self.stages = {}
        for name, stage in self.config.get('pipeline', {}).items():
            self.stages[name] = CrawlerStage(self, name, stage)

    def check_due(self):
        """Check if the last execution of this crawler is older than
        the scheduled interval."""
        if self.disabled:
            return False
        if self.is_running:
            return False
        if self.delta is None:
            return False
        last_run = self.last_run
        if last_run is None:
            return True
        now = datetime.utcnow()
        if now > last_run + self.delta:
            return True
        return False

    def flush(self):
        """Delete all run-time data generated by this crawler."""
        Queue.flush(self)
        Tag.delete(self)
        Event.delete(self)
        CrawlerState.flush(self)
        CrawlerRun.flush(self)

    def flush_events(self):
        Event.delete(self)

    def run(self, incremental=None, run_id=None):
        """Queue the execution of a particular crawler."""
        state = {
            'crawler': self.name,
            'run_id': run_id,
            'incremental': settings.INCREMENTAL
        }
        if incremental is not None:
            state['incremental'] = incremental

        # TaskRunner.execute(stage.name, state, {})
        Queue.queue(self.init_stage, state, {})

        if not settings.REDIS_HOST:
            TaskRunner.run()

    @property
    def is_running(self):
        """Is the crawler currently running?"""
        return Queue.is_running(self)

    @property
    def last_run(self):
        return CrawlerState.last_run(self)

    @property
    def op_count(self):
        """Total operations performed for this crawler"""
        return CrawlerState.op_count(self)

    @property
    def runs(self):
        return CrawlerState.runs(self)

    @property
    def latest_runid(self):
        return CrawlerState.latest_runid(self)

    def cancel(self):
        Queue.flush(self)

    def get(self, name):
        return self.stages.get(name)

    def __str__(self):
        return self.name

    def __iter__(self):
        return iter(self.stages.values())

    def __repr__(self):
        return '<Crawler(%s)>' % self.name



import requests
from pprint import pprint  # noqa
from banal import clean_dict
from six.moves.urllib.parse import urljoin

from memorious import settings
from memorious.model.common import dump_json


def aleph_emit(context, data):
    if not settings.ALEPH_HOST:
        context.log.warning("No $MEMORIOUS_ALEPH_HOST, skipping upload...")
        return
    if not settings.ALEPH_API_KEY:
        context.log.warning("No $MEMORIOUS_ALEPH_API_KEY, skipping upload...")
        return

    session = requests.Session()
    session.headers['Authorization'] = 'apikey %s' % settings.ALEPH_API_KEY
    collection_id = get_collection_id(context, session)
    if collection_id is None:
        context.log.warning("Cannot get aleph collection.")
        return

    content_hash = data.get('content_hash')
    source_url = data.get('source_url', data.get('url'))
    foreign_id = data.get('foreign_id', data.get('request_id', source_url))
    if context.skip_incremental(collection_id, foreign_id, content_hash):
        context.log.info("Skip aleph upload: %s", foreign_id)
        return

    meta = {
        'crawler': context.crawler.name,
        'foreign_id': foreign_id,
        'source_url': source_url,
        'title': data.get('title'),
        'author': data.get('author'),
        'file_name': data.get('file_name'),
        'retrieved_at': data.get('retrieved_at'),
        'modified_at': data.get('modified_at'),
        'published_at': data.get('published_at'),
        'headers': data.get('headers', {})
    }

    languages = context.params.get('languages')
    meta['languages'] = data.get('languages', languages)
    countries = context.params.get('countries')
    meta['countries'] = data.get('countries', countries)
    mime_type = context.params.get('mime_type')
    meta['mime_type'] = data.get('mime_type', mime_type)

    if data.get('parent_foreign_id'):
        meta['parent'] = {'foreign_id': data.get('parent_foreign_id')}

    meta = clean_dict(meta)
    # pprint(meta)

    url = make_url('collections/%s/ingest' % collection_id)
    label = meta.get('file_name', meta.get('source_url'))
    context.log.info("Upload: %s", label)
    with context.load_file(content_hash) as fh:
        if fh is None:
            return
        res = session.post(url,
                           data={'meta': dump_json(meta)},
                           files={'file': fh})
        if not res.ok:
            context.emit_warning("Error: %r" % res.text)
        else:
            document = res.json().get('documents')[0]
            context.log.info("Document ID: %s", document['id'])


def get_collection_id(context, session):
    url = make_url('collections')
    if hasattr(context.stage, '_aleph_cid'):
        return context.stage._aleph_cid
    foreign_id = context.get('collection', context.crawler.name)
    res = session.get(url, params={
        'filter:foreign_id': foreign_id
    })
    data = res.json()
    for coll in data.get('results'):
        if coll.get('foreign_id') == foreign_id:
            context.stage._aleph_cid = coll.get('id')
            return context.stage._aleph_cid

    res = session.post(url, json={
        'label': context.crawler.description,
        'casefile': False,
        'foreign_id': foreign_id
    })

    if not res.ok:
        context.log.info("Could not create collection: %s", foreign_id)
        return

    context.stage._aleph_cid = res.json().get('id')
    return context.stage._aleph_cid


def make_url(path):
    prefix = urljoin(settings.ALEPH_HOST, '/api/2/')
    return urljoin(prefix, path)



from memorious.operations.fetch import fetch, dav_index, session
from memorious.operations.parse import parse
from memorious.operations.clean import clean_html
from memorious.operations.aleph import aleph_emit
from memorious.operations.initializers import seed, sequence, dates
from memorious.operations.initializers import enumerate
from memorious.operations.debug import inspect
from memorious.operations.documentcloud import documentcloud_query
from memorious.operations.store import directory
from memorious.operations.extract import extract
from memorious.operations.db import db
from memorious.operations.ftp import ftp_fetch

__all__ = [fetch, parse, aleph_emit, seed, sequence, inspect, dates,
           documentcloud_query, dav_index, session, directory,
           enumerate, clean_html, extract, db, ftp_fetch]



from memorious.operations.documentcloud import documentcloud_query


class TestDocumentCloud(object):
    def test_query(self, context, mocker):
        data = {}
        context.params["query"] = "money"

        mocker.patch.object(context, "emit")
        mocker.patch.object(context, "recurse")

        documentcloud_query(context, data)
        assert context.emit.call_count == 100
        assert context.recurse.call_count == 1
        context.recurse.assert_called_once_with(data={'page': 2})



import numbers
import re
from normality import stringify
from dateutil.parser import parse


class ContextCheck(object):

    def __init__(self, context):
        self.context = context

    def shout(self, msg, strict=False, *args):
        if strict:
            raise ValueError(msg % args)
        else:
            self.context.log.info(msg, *args)

    def is_not_empty(self, value, strict=False):
        """if value is not empty"""
        value = stringify(value)
        if value is not None:
            return
        self.shout('Value %r is empty', strict, value)

    def is_numeric(self, value, strict=False):
        """if value is numeric"""
        value = stringify(value)
        if value is not None:
            if value.isnumeric():
                return
        self.shout('value %r is not numeric', strict, value)

    def is_integer(self, value, strict=False):
        """if value is an integer"""
        if value is not None:
            if isinstance(value, numbers.Number):
                return
        value = stringify(value)
        if value is not None and value.isnumeric():
            return
        self.shout('value %r is not an integer', strict, value)

    def match_date(self, value, strict=False):
        """if value is a date"""
        value = stringify(value)
        try:
            parse(value)
        except Exception:
            self.shout('Value %r is not a valid date', strict, value)

    def match_regexp(self, value, q, strict=False):
        """if value matches a regexp q"""
        value = stringify(value)
        mr = re.compile(q)
        if value is not None:
            if mr.match(value):
                return
        self.shout('%r not matching the regexp %r', strict, value, q)

    def has_length(self, value, q, strict=False):
        """if value has a length of q"""
        value = stringify(value)
        if value is not None:
            if len(value) == q:
                return
        self.shout('Value %r not matching length %r', strict, value, q)

    def must_contain(self, value, q, strict=False):
        """if value must contain q"""
        if value is not None:
            if value.find(q) != -1:
                return
        self.shout('Value %r does not contain %r', strict, value, q)



import random
import logging
from collections import deque
from datetime import datetime, timedelta

from memorious.core import manager
from memorious.model.common import Base, pack_datetime, unpack_datetime
from memorious.model.common import unpack_int, load_json, dump_json
from memorious.util import make_key

log = logging.getLogger(__name__)


class Queue(Base):
    """Manage the execution of tasks in the system."""

    @classmethod
    def serialize_task_data(cls, stage, state, data, delay):
        task_data = {}
        task_data["stage"] = stage
        task_data["state"] = state
        task_data["data"] = data
        if delay:
            if isinstance(delay, (int, float)):
                delay = timedelta(seconds=int(delay))

            if isinstance(delay, timedelta):
                delay = datetime.utcnow() + delay

            if isinstance(delay, datetime):
                task_data["next_allowed_exec_time"] = pack_datetime(delay)

        return dump_json(task_data)

    @classmethod
    def tasks(cls):
        queues = [make_key('queue', c, s) for c, s in manager.stages]
        random.shuffle(queues)
        while True:
            task_data_tuple = cls.conn.blpop(queues)
            # blpop blocks until it finds something. But fakeredis has no
            # blocking support. So it justs returns None.
            if not task_data_tuple:
                return

            key, json_data = task_data_tuple
            # Shift the queues list so that the matching key is at the
            # very end of the list, priorising all other crawlers.
            # queues = list(reversed(queues))
            deq = deque(queues)
            deq.rotate((queues.index(key) * -1) - 1)
            queues = list(deq)

            task_data = load_json(json_data)
            stage = task_data["stage"]
            state = task_data["state"]
            data = task_data["data"]
            next_time = task_data.get("next_allowed_exec_time")
            next_time = unpack_datetime(next_time)
            crawler = state.get('crawler')
            cls.conn.decr(make_key('queue_pending', crawler))
            yield (stage, state, data, next_time)

    @classmethod
    def queue(cls, stage, state, data, delay=None):
        crawler = state.get('crawler')
        task_data = cls.serialize_task_data(stage, state, data, delay)
        cls.conn.rpush(make_key('queue', crawler, stage), task_data)
        cls.conn.incr(make_key('queue_pending', crawler))

    @classmethod
    def size(cls, crawler):
        """Total operations pending for this crawler"""
        key = make_key('queue_pending', crawler)
        return unpack_int(cls.conn.get(key))

    @classmethod
    def is_running(cls, crawler):
        """Is the crawler currently running?"""
        if crawler.disabled:
            return False
        return cls.size(crawler) > 0

    @classmethod
    def flush(cls, crawler):
        prefix = make_key('queue', crawler, '*')
        for key in cls.conn.scan_iter(prefix):
            cls.conn.delete(key)
            cls.conn.ltrim(key, 0, -1)
            cls.conn.srem("queues_set", key)
        cls.conn.delete(make_key('queue_pending', crawler))



from setuptools import setup, find_packages

setup(
    name='memorious',
    version='0.7.4',
    description="A minimalistic, recursive web crawling library for Python.",
    long_description="",
    classifiers=[
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ],
    keywords='',
    author='Journalism Development Network',
    author_email='data@occrp.org',
    url='http://github.com/alephdata/memorious',
    license='MIT',
    packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
    namespace_packages=[],
    include_package_data=True,
    package_data={'memorious': ['migrate/*', 'migrate/versions/*']},
    zip_safe=False,
    install_requires=[
        'banal',
        'requests[security] >= 2.5',
        'requests_ftp',
        'click',
        'lxml >= 3',
        'PyYAML >= 3.10',
        'normality >= 0.5.7',
        'tabulate',
        'dataset >= 1.0.8',
        'six',
        'storagelayer >= 0.3.2',
        'urlnormalizer >= 1.2.0',
        'celestial >= 0.2.0',
        'werkzeug',
        'pycountry',
        'countrynames',
        'dateparser',
        'stringcase',
        'raven',
        'python-redis-rate-limit == 0.0.5',
        'redis == 2.10.6',
        'blinker == 1.4',
        'boto3 == 1.4.8',
        'fakeredis',
        'attrs',
    ],
    entry_points={
        'console_scripts': [
            'memorious = memorious.cli:main'
        ],
        'memorious.plugins': [
            'reporting = memorious.reporting:init'
        ]
    },
    extras_require={
        'dev': [
            'pytest',
            'pytest-env',
            'pytest-cov',
            'pytest-mock',
        ]
    }
)



import logging

from memorious.model import CrawlerState, CrawlerRun
from memorious.signals import operation_start
from memorious.signals import operation_end

log = logging.getLogger(__name__)


def log_operation_start(context):
    crawler = context.crawler
    stage = context.stage
    run_id = context.run_id
    CrawlerState.record_operation_start(crawler, stage)
    CrawlerRun.record_operation_start(crawler, run_id)


def log_operation_end(context):
    crawler = context.crawler
    run_id = context.run_id
    CrawlerState.record_operation_end(crawler)
    CrawlerRun.record_operation_end(crawler, run_id)


def init():
    operation_start.connect(log_operation_start)
    operation_end.connect(log_operation_end)
